//----------------------------------------------------------------------------
//
// File:        mkspch.cpp
// Date:        06-May-2001
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 2001-2002 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "option.hpp"

DBG_REGISTER ( __FILE__ );

struct sNode {
    int          p_length;
    const char  *phrase;
    sNode       *prev;
    sNode       *next;
    const UCHAR *data;
    int          d_length;
    bool         copy;
    int          phrase_offset;
    int          data_offset;
};

#if defined ( _MSC_VER )

    char *strndup ( const char *string, size_t length )
    {
        FUNCTION_ENTRY ( NULL, "strndup", false );

        size_t i = 0;
        while (( string [i] != '\0' ) && ( i < length )) {
            i++;
        }

        char *ptr = ( char * ) malloc ( i + 1 );
        memcpy ( ptr, string, i );
        ptr [i] = '\0';

        return ptr;
    }

#endif

const char *ReadPhrase ( const char *ptr, sNode *node )
{
    FUNCTION_ENTRY ( NULL, "ReadPhrase", false );

    if ( *ptr == '"' ) {
        const char *end = ptr + 1;
        while (( *end != '\0' ) && ( *end != '"' )) end++;
        node->p_length = end - ptr - 1;
        node->phrase   = node->p_length ? strndup ( ptr + 1, node->p_length ) : NULL;
        ptr = ( *end == '"' ) ? end + 1 : end;
    } else {
        const char *end = ptr + 1;
        while (( *end != '\0' ) && ! isspace ( *end )) end++;
        node->p_length = end - ptr;
        node->phrase   = node->p_length ? strndup ( ptr, node->p_length ) : NULL;
        ptr = end;
    }

    while ( isspace ( *ptr )) ptr++;

    return ptr;
}

void ReadHexData ( const char *ptr, sNode *node )
{
    FUNCTION_ENTRY ( NULL, "ReadHexData", false );

    UCHAR buffer [0x8000];
    int  data_length = 0;

    while ( *ptr ) {
        int byte  = 0;
        int count = sscanf ( ptr, "%02X", &byte );

        if ( count != 1 ) {
            fprintf ( stderr, "bad data!\n" );
            return;
        }

        buffer [data_length++] = ( UCHAR ) byte;

        ptr += 2;
        while ( isspace ( *ptr )) ptr++;
    }

    UCHAR *data = new UCHAR [ data_length ];
    memcpy ( data, buffer, data_length );

    node->data     = data;
    node->d_length = data_length;
}

static UCHAR *vsmDataPtr;
static int    vsmBitsLeft;
static int    vsmData;

void WriteBits ( int data, int count )
{
    FUNCTION_ENTRY ( NULL, "WriteBits", false );

    int mask = 1 << ( count - 1 );

    if ( data >= ( 1 << count )) {
        WARNING ( "data (" << data << ") contains more than " << count << " bits" );
    }

    while ( count-- ) {

        vsmBitsLeft--;
        vsmData <<= 1;

        if ( data & mask ) {
            vsmData |= 1;
        }

        mask >>= 1;

        if ( vsmBitsLeft == 0 ) {
            *vsmDataPtr++ = ( UCHAR ) vsmData;
            vsmBitsLeft   = 8;
        }
    }
}

const char *ReadCoefficient ( const char *ptr, int bits )
{
    FUNCTION_ENTRY ( NULL, "ReadCoefficient", false );

    if ( *ptr != 'K' ) throw "Syntax error";

    int coefficient = atoi ( ptr + 1 );

    WriteBits ( coefficient, bits );

    while ( ! isspace ( *ptr )) ptr++;
    while ( isspace ( *ptr )) ptr++;

    return ptr;
}

const char *ParseFrame ( const char *ptr )
{
    FUNCTION_ENTRY ( NULL, "ParseFrame", false );

    switch ( *ptr ) {
        case 'Z' :
            ptr++;
            WriteBits ( 0x00, 4 );
            break;
        case 'S' :
            ptr++;
            WriteBits ( 0x0F, 4 );
            break;
        case 'P' :
            {
                int pitch = atoi ( ptr + 1 );

                while ( ! isspace ( *ptr )) ptr++;
                while ( isspace ( *ptr )) ptr++;

                if ( *ptr != 'E' ) throw "Syntax error";

                int energy = atoi ( ptr + 1 );

                while ( ! isspace ( *ptr )) ptr++;
                while ( isspace ( *ptr )) ptr++;

                WriteBits ( energy, 4 );

                if ( *ptr == 'R' ) {
                    ptr++;
                    WriteBits ( 1, 1 );
                    WriteBits ( pitch, 6 );
                    break;
                } else {
                    WriteBits ( 0, 1 );
                    WriteBits ( pitch, 6 );

                    ptr = ReadCoefficient ( ptr, 5 );
                    ptr = ReadCoefficient ( ptr, 5 );
                    ptr = ReadCoefficient ( ptr, 4 );
                    ptr = ReadCoefficient ( ptr, 4 );

                    if ( pitch != 0 ) {
                        ptr = ReadCoefficient ( ptr, 4 );
                        ptr = ReadCoefficient ( ptr, 4 );
                        ptr = ReadCoefficient ( ptr, 4 );
                        ptr = ReadCoefficient ( ptr, 3 );
                        ptr = ReadCoefficient ( ptr, 3 );
                        ptr = ReadCoefficient ( ptr, 3 );
                    }

                }
            }
            break;
        default :
            throw "Error parsing speech frame";
            break;
    }

    while ( isspace ( *ptr )) ptr++;

    return ptr;
}

void ReadParsedData ( const char *ptr, sNode *node )
{
    FUNCTION_ENTRY ( NULL, "ReadParsedData", false );

    try {

        UCHAR buffer [0x8000];

        vsmDataPtr  = buffer;
        vsmBitsLeft = 8;
        vsmData     = 0;

        while ( *ptr ) {
            ptr = ParseFrame ( ptr );
        }

        if ( vsmBitsLeft != 8 ) {
            *vsmDataPtr++ = ( UCHAR ) ( vsmData << vsmBitsLeft );
        }

        int data_length = vsmDataPtr - buffer;

        UCHAR *data = new UCHAR [ data_length ];
        memcpy ( data, buffer, data_length );

        node->data     = data;
        node->d_length = data_length;
    }

    catch ( const char *msg ) {
        fprintf ( stderr, "Phrase: \"%*.*s\" - %s\n", node->p_length, node->p_length, node->phrase, msg );
    }
}

void ReadData ( const char *ptr, sNode *node )
{
    FUNCTION_ENTRY ( NULL, "ReadData", false );

    // We'll figure this out later
    node->copy     = false;
    node->data     = NULL;
    node->d_length = 0;

    if ( strncmp ( ptr, "<null>", 6 ) == 0 ) {
        UCHAR *data = new UCHAR [1];

        data [0] = 0xF0;

        node->data     = data;
        node->d_length = 1;
    } else if ( *ptr == '-' ) {
        ptr += 1;
        while ( isspace ( *ptr )) ptr++;
        if ( isxdigit ( *ptr )) {
            ReadHexData ( ptr, node );
        } else {
            ReadParsedData ( ptr, node );
        }
    }
}

sNode *ReadNode ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ReadNode", true );

    static int line = 0;

    static char buffer [4096];
    const char *ptr;

    // Read the next line from the file - ignore blank or comment lines
    do {
        line++;
        if ( fgets ( buffer, sizeof ( buffer ), file ) == 0 ) return NULL;
        ptr = buffer;
        while ( isspace ( *ptr )) ptr++;
    } while (( *ptr == '#' ) || ( *ptr == '\0' ));

    // Remove the trailing '\n'
    buffer [ strlen ( buffer ) - 1 ] = '\0';

    sNode *node = new sNode;
    memset ( node, 0, sizeof ( sNode ));

    ptr = ReadPhrase ( ptr, node );

    if ( node->phrase == NULL ) {
        fprintf ( stdout, "Invalid phrase on line %d\n", line );
        exit ( -1 );
    }

    ReadData ( ptr, node );

    if ( node->data == NULL ) {
        fprintf ( stdout, "Invalid speech data on line %d\n", line );
        exit ( -1 );
    }

    return node;
}

int sort_node ( const void *ptr1, const void *ptr2 )
{
    FUNCTION_ENTRY ( NULL, "sort_node", false );

    const sNode *node1 = * ( const sNode ** ) ptr1;
    const sNode *node2 = * ( const sNode ** ) ptr2;

    return strcmp ( node1->phrase, node2->phrase );
}

sNode *SplitList ( sNode **list, int size )
{
    FUNCTION_ENTRY ( NULL, "SplitList", true );

    if ( size == 0 ) return NULL;

    int rootIndex = size / 2;
    sNode *root = list [ rootIndex ];

    root->prev = SplitList ( list, rootIndex );
    root->next = SplitList ( list + rootIndex + 1, size - rootIndex - 1 );

    return root;
}

void CalculateOffsets ( sNode *root, sNode **list, int count )
{
    FUNCTION_ENTRY ( NULL, "CalculateOffsets", true );

    // Calculate the phrase offsets
    int offset = 1;
    root->phrase_offset = offset;
    offset += 1 + strlen ( root->phrase ) + 2 + 2 + 4;
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        if ( node == root ) continue;
        node->phrase_offset = offset;
        offset += 1 + strlen ( node->phrase ) + 2 + 2 + 4;
    }

    // Calculate the data offsets
    for ( int i = 0; i < count; i++ ) {
        sNode *node = list [i];
        // Look for a match in an earlier phrase
        for ( int j = 0; j < i; j++ ) {
            if ( list [j]->d_length >= node->d_length ) {
                // Look for this phrase at the end of a previous one
                int offset = list [j]->d_length - node->d_length;
                if ( memcmp (( char * ) list [j]->data + offset, node->data, node->d_length ) == 0 ) {
                    node->copy = true;
                    node->data_offset = list [j]->data_offset;
                }
            }
        }
        if ( node->copy == false ) {
            node->data_offset = offset;
            offset += node->d_length;
        }
    }
}

UCHAR *StoreNode ( const sNode *node, UCHAR *ptr )
{
    FUNCTION_ENTRY ( NULL, "StoreNode", true );

    ptr += sprintf (( char * ) ptr, "%c%s", node->p_length, node->phrase );

    int offset;

    offset = node->prev ? node->prev->phrase_offset : 0;
    *ptr++ = ( UCHAR ) ( offset >> 8 );
    *ptr++ = ( UCHAR ) ( offset & 0xFF );

    offset = node->next ? node->next->phrase_offset : 0;
    *ptr++ = ( UCHAR ) ( offset >> 8 );
    *ptr++ = ( UCHAR ) ( offset & 0xFF );

    offset = node->data_offset;
    *ptr++ = ( UCHAR ) ( '\0' );              // Dunno what this is yet
    *ptr++ = ( UCHAR ) ( offset >> 8 );
    *ptr++ = ( UCHAR ) ( offset & 0xFF );
    *ptr++ = ( UCHAR ) ( node->d_length );

    return ptr;
}

void MakeROM ( const sNode *root, sNode **list, int count, UCHAR *rom )
{
    FUNCTION_ENTRY ( NULL, "MakeROM", true );

    UCHAR *ptr = rom;

    *ptr++ = 0xAA;

    // Store the phrases
    ptr = StoreNode ( root, ptr );
    for ( int i = 0; i < count; i++ ) {
        const sNode *node = list [i];
        if ( node == root ) continue;
        ptr = StoreNode ( node, ptr );
    }

    // Store the data
    for ( int i = 0; i < count; i++ ) {
        const sNode *node = list [i];
        if ( node->copy == false ) {
            memcpy ( ptr, node->data, node->d_length );
            ptr += node->d_length;
        }
    }
}

bool ParseFileName ( const char *arg, void *filename )
{
    FUNCTION_ENTRY ( NULL, "ParseFileName", true );

    const char *ptr = strchr ( arg, '=' );

    // Handle special case of extract all
    if ( ptr == NULL ) {
        fprintf ( stderr, "A filename needs to be specified: '%s'\n", arg );
        return false;
    }

    strcpy (( char * ) filename, ptr + 1 );

    return true;
}

void PrintUsage ()
{
    FUNCTION_ENTRY ( NULL, "PrintUsage", true );

    fprintf ( stdout, "Usage: mkspch [options] file\n" );
    fprintf ( stdout, "\n" );
}

int main ( int argc, char *argv[] )
{
    FUNCTION_ENTRY ( NULL, "main", true );

    char outputFile [256] = "spchrom.bin";

    sOption optList [] = {
        { 'o', "output=*<filename>",  OPT_NONE,                      true, outputFile,    ParseFileName,  "Create output file <filename>" },
        { 'v', "verbose",             OPT_VALUE_SET | OPT_SIZE_BOOL, true, &verbose,      NULL,           "Display extra information" }
    };

    if ( argc == 1 ) {
        PrintHelp ( SIZE ( optList ), optList );
        return 0;
    }

    printf ( "TI-99/4A Speech ROM Utility\n" );

    int index = 1;
    index = ParseArgs ( index, argc, argv, SIZE ( optList ), optList );

    if ( index >= argc ) {
        fprintf ( stderr, "No input file specified\n" );
        return -1;
    }

    FILE *datFile = fopen ( argv [index], "rt" );
    if ( datFile == NULL ) {
        fprintf ( stderr, "Unable to open input file \"%s\"\n", argv [index] );
        return -1;
    }

    sNode *list [ 3277 ];
    int count = 0;

    // Read in the basic phrases and their speech data
    for ( EVER ) {
        sNode *node = ReadNode ( datFile );
        if ( node == NULL ) break;
        list [count++] = node;
    }

    fclose ( datFile );

    fprintf ( stdout, "\n" );
    fprintf ( stdout, "%5d phrases read\n", count );
    fprintf ( stdout, "\n" );

    // Put them in alphabetical order
    qsort ( list, count, sizeof ( sNode * ), sort_node );

    // Build the binary tree
    sNode *root = SplitList ( list, count );

    CalculateOffsets ( root, list, count );

    UCHAR ROM [ 0x8000 ];
    memset ( ROM, 0, sizeof ( ROM ));

    MakeROM ( root, list, count, ROM );

    for ( int i = 0; i < count; i++ ) {
        free (( void * ) list [i]->phrase );
        delete [] ( UCHAR * ) list [i]->data;
        delete list [i];
    }

    FILE *romFile = fopen ( outputFile, "wb" );
    if ( romFile == NULL ) {
        fprintf ( stderr, "Unable to open output file \"%s\"\n", outputFile );
        return -1;
    }

    fwrite ( ROM, sizeof ( ROM ), 1, romFile );

    fclose ( romFile );

    return 0;
}
